import datetime
from decimal import Decimal
from typing import Optional, Any, Dict, List

from async_lru import alru_cache
from dateutil.relativedelta import relativedelta

from sirius import common
from sirius.common import Currency
from sirius.constants import SiriusEnvironmentSecretKey
from sirius.http_requests import HTTPResponse, AsyncHTTPSession
from sirius.market_data import Stock, StockMarketData, Option, OptionType, OptionMarketData

BASE_URL: str = "https://api.massive.com"


class MassiveStock(Stock):

    @staticmethod
    @alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
    async def _get_raw_ticker_search_data(ticker: str) -> List[Dict[str, Any]]:
        api_key: str = await common.get_environmental_secret(SiriusEnvironmentSecretKey.MASSIVE_API_KEY)
        response: HTTPResponse = await AsyncHTTPSession(BASE_URL).get(f"{BASE_URL}/v3/reference/tickers", query_params={
            "ticker": ticker,
            "market": "stocks",
            "apikey": api_key
        })

        return list(filter(lambda r: r["currency_name"] == "usd", response.data["results"]))

    @staticmethod
    async def _find(ticker: str) -> Optional["Stock"]:
        raw_data_list: List[Dict[str, Any]] = await MassiveStock._get_raw_ticker_search_data(ticker)
        if not raw_data_list:
            return None

        raw_data: Dict[str, Any] = raw_data_list[0]
        return MassiveStock(
            name=raw_data["name"],
            currency=Currency(raw_data["currency_name"].upper()),
            ticker=raw_data["ticker"]
        )


class MassiveStockMarketData(StockMarketData):

    @staticmethod
    @alru_cache(maxsize=50, ttl=43_200)  # 12 hour cache
    async def _get_raw_time_series_data(ticker: str) -> Dict[datetime.datetime, Any]:
        api_key: str = await common.get_environmental_secret(SiriusEnvironmentSecretKey.MASSIVE_API_KEY)
        from_date_str: str = (datetime.datetime.now() - relativedelta(years=5)).strftime("%Y-%m-%d")
        to_date_str: str = datetime.datetime.now().strftime("%Y-%m-%d")
        response: HTTPResponse = await AsyncHTTPSession(BASE_URL).get(f"{BASE_URL}/v2/aggs/ticker/{ticker}/range/1/day/{from_date_str}/{to_date_str}", query_params={
            "adjusted": "true",
            "limit": 50000,
            "apikey": api_key
        })

        return {
            datetime.datetime.fromtimestamp(data["t"] / 1000): data
            for data in response.data["results"]
        }

    @staticmethod
    async def _get(abstract_stock: Stock, from_datestamp: datetime.date, to_datestamp: datetime.date) -> List["StockMarketData"]:
        raw_data: Dict[datetime.datetime, Any] = await MassiveStockMarketData._get_raw_time_series_data(abstract_stock.ticker)
        return [MassiveStockMarketData(
            open=Decimal(raw_data[timestamp]["o"]),
            high=Decimal(raw_data[timestamp]["h"]),
            low=Decimal(raw_data[timestamp]["l"]),
            close=Decimal(raw_data[timestamp]["c"]),
            timestamp=timestamp,
            stock=abstract_stock
        ) for timestamp in raw_data.keys()
            if from_datestamp <= timestamp.date() <= to_datestamp]

    @staticmethod
    @alru_cache(maxsize=50, ttl=60)  # 60 second cache
    async def _get_raw_latest_stock_market_data(ticker: str) -> Dict[str, Any]:
        api_key: str = await common.get_environmental_secret(SiriusEnvironmentSecretKey.MASSIVE_API_KEY)
        response: HTTPResponse = await AsyncHTTPSession(BASE_URL).get(f"{BASE_URL}/v2/snapshot/locale/us/markets/stocks/tickers", query_params={
            "tickers": ticker,
            "apikey": api_key
        })

        return response.data["tickers"][0]

    @staticmethod
    async def get_latest(abstract_stock: Stock) -> "MassiveStockMarketData":
        raw_data: Dict[str, Any] = await MassiveStockMarketData._get_raw_latest_stock_market_data(abstract_stock.ticker)
        is_latest_market_data: bool = raw_data["min"]["t"] != 0

        if raw_data["min"]["t"] != 0:
            raw_data_2: Dict[str, int] = raw_data["min"]
            timestamp: datetime.datetime = datetime.datetime.fromtimestamp(raw_data["updated"] / 1_000_000_000)
        else:
            raw_data_2 = raw_data["prevDay"]
            previous_business_day: datetime.date = common.get_previous_business_day(datetime.datetime.now())
            timestamp = datetime.datetime.combine(previous_business_day, datetime.time(23, 59, 59))

        return MassiveStockMarketData(
            open=Decimal(raw_data_2["o"]),
            high=Decimal(raw_data_2["h"]),
            low=Decimal(raw_data_2["l"]),
            close=Decimal(raw_data_2["c"]),
            timestamp=timestamp,
            stock=abstract_stock
        )


class MassiveOption(Option):

    @staticmethod
    async def _find(option_id: str) -> Optional["Option"]:
        ticker: str = option_id.split(" | ")[0]
        option_list: List[MassiveOption] = await MassiveOption._find_all_from_ticker(ticker)
        option_map: Dict[str, MassiveOption] = {option.id: option for option in option_list}
        return option_map.get(option_id)

    @staticmethod
    @alru_cache(maxsize=50, ttl=60)  # 60 second cache
    async def _get_raw_data(ticker: str) -> List[Dict[str, Any]]:
        api_key: str = await common.get_environmental_secret(SiriusEnvironmentSecretKey.MASSIVE_API_KEY)
        response: HTTPResponse = await AsyncHTTPSession(BASE_URL).get(f"{BASE_URL}/v3/snapshot/options/{ticker}", query_params={
            "limit": "250",
            "apikey": api_key
        })

        all_data: List[Dict[str, Any]] = response.data["results"]
        while "next_url" in response.data:
            response = await AsyncHTTPSession(BASE_URL).get(f"{response.data["next_url"]}&apikey={api_key}")
            all_data = all_data + response.data["results"]

        return all_data

    @staticmethod
    async def _find_all_from_ticker(ticker: str) -> List["MassiveOption"]:
        option_list: List[MassiveOption] = []
        underlying_stock: Stock = await Stock.find(ticker)
        if not underlying_stock:
            return []

        for data in await MassiveOption._get_raw_data(ticker):
            expiry_date: datetime.date = datetime.datetime.strptime(data["details"]["expiration_date"], "%Y-%m-%d").date()
            if expiry_date < datetime.datetime.now().date():
                continue

            strike_price: Decimal = Decimal(data["details"]["strike_price"])
            option_type: OptionType = OptionType(str(data["details"]["contract_type"]).upper())
            option_id: str = Option._get_id(ticker, expiry_date, strike_price, option_type)
            option_list.append(MassiveOption(
                id=option_id,
                underlying_stock=underlying_stock,
                strike_price=strike_price,
                expiry_date=expiry_date,
                type=option_type,
                currency=underlying_stock.currency,
                name=option_id
            ))

        return option_list

    @staticmethod
    async def _find_all_options(ticker: str, number_of_days_to_expiry: int) -> List["Option"]:
        data = await MassiveOption._find_all_from_ticker(ticker)
        return [option for option in data
                if (option.expiry_date - datetime.datetime.now().date()).days == number_of_days_to_expiry]


class MassiveOptionMarketData(OptionMarketData):

    @staticmethod
    @alru_cache(maxsize=50, ttl=60)  # 60 second cache
    async def _get_raw_data(ticker: str) -> Dict[str, Any] | None:
        api_key: str = await common.get_environmental_secret(SiriusEnvironmentSecretKey.MASSIVE_API_KEY)
        from_date_str: str = (datetime.datetime.now() - relativedelta(days=5)).strftime("%Y-%m-%d")
        to_date_str: str = datetime.datetime.now().strftime("%Y-%m-%d")
        response: HTTPResponse = await AsyncHTTPSession(BASE_URL).get(f"{BASE_URL}/v2/aggs/ticker/O:{ticker}/range/1/second/{from_date_str}/{to_date_str}", query_params={
            "limit": "1",
            "adjusted": "true",
            "sort": "desc",
            "apikey": api_key
        })

        return response.data["results"][0] if "results" in response.data else None


    @staticmethod
    async def _get_latest(abstract_option: Option) -> "MassiveOptionMarketData":
        raw_data: Dict[str, Any] = await MassiveOptionMarketData._get_raw_data(abstract_option.id)
        last: Decimal = Decimal(raw_data["l"] if raw_data else "0")
        return MassiveOptionMarketData(option=abstract_option, last=last)



BASE_URL: str = "https://discord.com/api/"
VERSION_NUMBER: str = "v10"
URL: str = f"{BASE_URL}/{VERSION_NUMBER}/"

ENDPOINT__BOT__GET_BOT: str = f"{URL}/users/@me"
ENDPOINT__SERVER__GET_ALL_SERVERS: str = f"{URL}/users/@me/guilds"
ENDPOINT__SERVER__GET_ALL_USERS: str = f"{URL}/guilds/$serverID/members"
ENDPOINT__SERVER__GET_ALL_ROLES: str = f"{URL}/guilds/$serverID/roles"
ENDPOINT__CHANNEL__CREATE_CHANNEL_OR_GET_ALL_CHANNELS: str = f"{URL}/guilds/$serverID/channels"
ENDPOINT__CHANNEL__SEND_MESSAGE: str = f"{URL}/channels/$channelID/messages"
ENDPOINT__CHANNEL__DELETE: str = f"{URL}/channels/$channelID"



import datetime
from decimal import Decimal
from enum import Enum
from typing import List, Dict, Any

from pydantic import PrivateAttr

from sirius import common
from sirius.common import DataClass, Currency
from sirius.constants import SiriusEnvironmentSecretKey
from sirius.http_requests import AsyncHTTPSession, HTTPResponse
from sirius.wise import constants


async def get_http_session(api_key: str) -> AsyncHTTPSession:
    api_key = await common.get_environmental_secret(SiriusEnvironmentSecretKey.WISE_API_KEY, "PLACEHOLDER_WISE_API_KEY") if api_key is None else api_key
    return AsyncHTTPSession(constants.URL, {"Authorization": f"Bearer {api_key}"})


class AccountType(Enum):
    STANDARD = "STANDARD"
    SAVINGS = "SAVINGS"


class ProfileType(Enum):
    PERSONAL = "PERSONAL"
    BUSINESS = "BUSINESS"


class Transaction(DataClass):
    amount: Decimal
    currency: Currency
    description: str
    running_balance: Decimal


class Account(DataClass):
    id: int
    profile_id: int
    currency: Currency
    type: AccountType
    name: str
    balance: Decimal
    _http_session: AsyncHTTPSession = PrivateAttr()

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._http_session = data["http_session"]

    async def get_transactions(self, from_time: datetime.datetime | None = None, to_time: datetime.datetime | None = None, number_of_past_hours: int | None = None) -> List[Transaction]:
        transaction_list: List[Transaction] = []
        number_of_past_hours = number_of_past_hours if number_of_past_hours else 24
        from_time = from_time if from_time else datetime.datetime.now() - datetime.timedelta(hours=number_of_past_hours)
        to_time = to_time if to_time else datetime.datetime.now()
        url: str = constants.ENDPOINT__BALANCE__GET_TRANSACTIONS.replace("$profileId", str(self.profile_id)).replace("$balanceId", str(self.id))
        query_params: Dict[str, Any] = {
            "currency": self.currency.value,
            "intervalStart": f"{from_time.astimezone(datetime.timezone.utc).replace(microsecond=0).isoformat().split('+')[0]}Z",
            "intervalEnd": f"{to_time.astimezone(datetime.timezone.utc).replace(microsecond=0).isoformat().split('+')[0]}Z",
            "type": "COMPACT"
        }

        response: HTTPResponse = await self._http_session.get(url, query_params)
        for data in response.data["transactions"]:
            transaction = Transaction(
                amount=Decimal(str(data["amount"]["value"])),
                currency=Currency(data["amount"]["currency"]),
                description=data["details"]["description"],
                running_balance=Decimal(str(data["runningBalance"]["value"]))
            )
            transaction_list.append(transaction)

        return transaction_list


class Profile(DataClass):
    id: int
    type: ProfileType
    _account_list: List[Account] = []
    _http_session: AsyncHTTPSession = PrivateAttr()

    @property
    async def account_list(self) -> List[Account] | None:
        if len(self._account_list) != 0:
            return self._account_list

        url: str = constants.ENDPOINT__ACCOUNT__GET_ALL.replace("$profileId", str(self.id))
        response: HTTPResponse = await self._http_session.get(url)
        for data in response.data:
            account = Account(
                id=data["id"],
                profile_id=self.id,
                currency=Currency(data["totalWorth"]["currency"]),
                type=AccountType(data["type"]),
                name=data["name"] if data["name"] else f"{data['totalWorth']['currency']} Bank Account",
                balance=Decimal(str(data["totalWorth"]["value"])),
                http_session=self._http_session
            )
            self._account_list.append(account)

        return self._account_list

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._http_session = data["http_session"]


class WiseAccount(DataClass):
    personal_profile: Profile
    _http_session: AsyncHTTPSession
    business_profile: Profile | None = None

    @staticmethod
    async def get(api_key: str | None = None) -> "WiseAccount":
        http_session: AsyncHTTPSession = await get_http_session(api_key)
        response: HTTPResponse = await http_session.get(constants.ENDPOINT__PROFILE__GET_ALL)
        profile_list: List[Profile] = [Profile(id=data["id"], type=ProfileType(data["type"]), http_session=http_session) for data in response.data]
        personal_profile: Profile = next(filter(lambda p: p.type == ProfileType.PERSONAL, profile_list))
        business_profile: Profile = next(filter(lambda p: p.type == ProfileType.BUSINESS, profile_list), None)

        return WiseAccount(personal_profile=personal_profile, business_profile=business_profile, _http_session=http_session)



import asyncio
import datetime
import itertools
from decimal import Decimal
from typing import List, Dict, Any, cast, Set, Callable, Optional
from zoneinfo import ZoneInfo

import httpx
from async_lru import alru_cache
from pydantic import ConfigDict, SerializeAsAny

from sirius import common
from sirius.common import Currency, DataClass
from sirius.constants import SiriusEnvironmentSecretKey
from sirius.exceptions import SiriusException
from sirius.http_requests import HTTPResponse, ServerSideException, AsyncHTTPSession
from sirius.market_data import Stock, Option, StockMarketData, Exchange, OptionType, OptionMarketData

OPTIONS_DATE_FORMAT: str = "%b%y"
_session: AsyncHTTPSession | None = None


async def get_base_url() -> str:
    return await common.get_environmental_secret(SiriusEnvironmentSecretKey.IBKR_SERVICE_BASE_URL, "http://ibkr-gateway/")


async def get_session() -> AsyncHTTPSession:
    global _session
    if not _session:
        base_url: str = await get_base_url()
        _session = AsyncHTTPSession(base_url)
        _session.client = httpx.AsyncClient(verify=False, timeout=60)

    return _session


class IBKRException(SiriusException):
    pass


class IBKRContract(DataClass):
    contract_id: int

    @staticmethod
    @alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
    async def find_by_contract_id(contract_id: int) -> Optional["IBKRContract"]:
        session: AsyncHTTPSession = await get_session()
        response: HTTPResponse = await session.get(f"{await get_base_url()}iserver/contract/{contract_id}/info")
        instrument_type: str = response.data["instrument_type"]
        symbol: str = response.data["local_symbol"].replace(" ", "")

        if instrument_type == "STK":
            ibkr_contract: IBKRContract | None = cast(IBKRContract, await IBKRStock._find(symbol))
        elif instrument_type == "OPT":
            ibkr_contract = cast(IBKRContract, await IBKROption._find(symbol))
        else:
            return None

        assert ibkr_contract.contract_id == contract_id if ibkr_contract else True
        return ibkr_contract


class IBKRStock(Stock, IBKRContract):
    model_config = ConfigDict(frozen=True)

    @staticmethod
    @alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
    async def _find(ticker: str) -> Optional[Stock]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        response: HTTPResponse = await session.get(f"{base_url}iserver/secdef/search?symbol={ticker}&secType=STK")
        filtered_list: List[Dict[str, Any]] = list(filter(lambda d: d["description"] in Exchange, response.data))
        contract_id_list: List[int] = [int(data["conid"]) for data in filtered_list]

        if not contract_id_list:
            return None

        if len(contract_id_list) > 1:
            raise IBKRException(f"More than one stock found for the ticker: {ticker}")

        contract_id: int = contract_id_list[0]
        response = await session.get(f"{base_url}iserver/contract/{contract_id}/info")
        return IBKRStock(
            name=response.data["company_name"],
            currency=Currency(response.data["currency"]),
            ticker=response.data["symbol"],
            contract_id=contract_id
        )


class IBKROption(Option, IBKRContract):

    @staticmethod
    async def _find(option_id: str) -> Optional["Option"]:
        ticker: str = "".join(itertools.takewhile(str.isalpha, option_id))
        date_string: str = option_id.replace(ticker, "")[:6]
        expiry_date: datetime.date = datetime.datetime.strptime(date_string, "%y%m%d").date()
        strike_price: Decimal = Decimal(option_id[-8:]) / Decimal("1000")
        stock: Optional[IBKRStock] = cast(Optional[IBKRStock], await IBKRStock._find(ticker))
        option_type: OptionType = OptionType.CALL if option_id[option_id.find(date_string) + len(date_string)] == "C" else OptionType.PUT

        if not stock:
            raise IBKRException(f"Could not find the underlying stock for the option with ticker: {option_id}")

        all_option_list: List[IBKROption] = await IBKROption.__get_for_strike_and_expiry(stock, expiry_date, strike_price)
        try:
            return next(filter(lambda o: o.type == option_type and o.expiry_date == expiry_date, all_option_list))
        except StopIteration:
            raise IBKRException(f"Could not find option with ID: {option_id}")

    @staticmethod
    async def __get_all_expiry_month_list(stock: IBKRStock, number_of_days_to_expiry: int) -> List[datetime.date]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        response: HTTPResponse = await session.get(f"{base_url}iserver/secdef/search?symbol={stock.ticker}&secType=STK")
        data: Dict[str, Any] = next(filter(lambda c: int(c["conid"]) == stock.contract_id, response.data))
        option_data: Dict[str, Any] = next(filter(lambda o: o["secType"] == "OPT", data["sections"]))
        all_expiry_month_str_list: List[str] = option_data["months"].split(";")
        all_expiry_month_list: List[datetime.date] = [datetime.datetime.strptime(expiry_month, OPTIONS_DATE_FORMAT).date() for expiry_month in all_expiry_month_str_list]

        return [date for date in all_expiry_month_list if (date - datetime.datetime.now().date()).days <= number_of_days_to_expiry]

    @staticmethod
    async def __get_for_strike_and_expiry(stock: IBKRStock, expiry_month: datetime.date, strike_price: Decimal) -> List["IBKROption"]:
        option_list: List[IBKROption] = []
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        expiry_month_str: str = expiry_month.strftime(OPTIONS_DATE_FORMAT).upper()
        response: HTTPResponse = await session.get(
            f"{base_url}iserver/secdef/info",
            query_params={
                "conid": stock.contract_id,
                "sectype": "OPT",
                "month": expiry_month_str,
                "strike": float(strike_price)}
        )

        for data in response.data:
            expiry_date: datetime.date = datetime.datetime.strptime(data["maturityDate"], '%Y%m%d').date()
            option_type: OptionType = OptionType.CALL if data["right"] == "C" else OptionType.PUT
            option_id: str = Option._get_id(stock.ticker, expiry_date, strike_price, option_type)

            option_list.append(IBKROption(
                id=option_id,
                contract_id=data["conid"],
                strike_price=strike_price,
                expiry_date=expiry_date,
                type=option_type,
                underlying_stock=stock,
                name=option_id,
                currency=stock.currency
            ))

        return option_list

    @staticmethod
    @alru_cache(maxsize=50, ttl=43_200)  # 12 hour cache
    async def _find_all_options(ticker: str, number_of_days_to_expiry: int) -> List[Option]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        abstract_stock: Stock = await Stock.get(ticker)
        stock: IBKRStock = cast(IBKRStock, await IBKRStock._get_local_object(abstract_stock))
        option_contract_list: List[Option] = []
        expiry_month_list: List[datetime.date] = await IBKROption.__get_all_expiry_month_list(stock, number_of_days_to_expiry)
        expiry_month_str_list: List[str] = [expiry_month.strftime(OPTIONS_DATE_FORMAT).upper() for expiry_month in expiry_month_list]

        responses: List[HTTPResponse] = await asyncio.gather(*[
            session.get(f"{base_url}iserver/secdef/strikes", query_params={"conid": stock.contract_id, "sectype": "OPT", "month": expiry_month_str})
            for expiry_month_str in expiry_month_str_list
        ])
        for expiry_month, response in zip(expiry_month_list, responses):
            all_strike_price_set: Set[Decimal] = set([Decimal(str(strike_price)) for strike_price in response.data.get("call", [])])
            all_strike_price_set.update([Decimal(str(strike_price)) for strike_price in response.data.get("put", [])])
            all_option_contract_list: List[IBKROption] = list(itertools.chain.from_iterable(await asyncio.gather(*[
                IBKROption.__get_for_strike_and_expiry(stock, expiry_month, strike_price)
                for strike_price in all_strike_price_set
            ])))

            option_contract_list.extend(
                [option
                 for option in all_option_contract_list
                 if (option.expiry_date - datetime.datetime.now().date()).days == number_of_days_to_expiry]
            )

        return option_contract_list


class IBKRStockMarketData(StockMarketData):

    @staticmethod
    @alru_cache(maxsize=50, ttl=43_200)  # 12 hour cache
    async def _get_raw_ohlc_data(contract_id: int, from_timestamp: datetime.datetime, to_timestamp: datetime.datetime) -> List[Dict[str, Any]]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        DATE_FORMAT: str = "%Y%m%d-%H:%M:%S"
        try:
            response = await session.get(
                f"{base_url}iserver/marketdata/history",
                query_params={
                    "conid": contract_id,
                    "period": "999d",
                    "bar": "1d",
                    "startTime": (to_timestamp + datetime.timedelta(days=1)).strftime(DATE_FORMAT),  # IBKR sends data 1 day earlier, no idea why.
                    "direction": "-1"
                }
            )
        except ServerSideException as e:
            raise ServerSideException("Did not retrieve any historical market data due to: " + str(e))

        data = response.data.get("data", [])
        if not data:
            return []

        earliest_data_timestamp = min([datetime.datetime.fromtimestamp(d["t"] / 1000) for d in data])

        if earliest_data_timestamp.timestamp() > from_timestamp.timestamp():
            more_data = await IBKRStockMarketData._get_raw_ohlc_data(contract_id, from_timestamp, earliest_data_timestamp)
            return list(itertools.chain(data, more_data))

        return data

    @staticmethod
    @alru_cache(maxsize=50, ttl=60)  # 60 second cache
    async def _get_raw_latest_ohlc_data(contract_id: int) -> Dict[str, Any]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        is_response_valid: Callable = lambda r: "31" in r.data[0].keys() and r.data[0]["31"] != ""
        number_of_tries: int = 1
        response: HTTPResponse = await session.get(f"{base_url}iserver/marketdata/snapshot", query_params={"conids": contract_id, "fields": "7295,70,71,31,87"})
        while number_of_tries < 5 and not is_response_valid(response):
            await asyncio.sleep(0.1)
            response = await session.get(f"{base_url}iserver/marketdata/snapshot", query_params={"conids": contract_id, "fields": "7295,70,71,31,87"})
            number_of_tries = number_of_tries + 1

        if not is_response_valid(response):
            raise ServerSideException("Did not retrieve any market data.")

        return response.data[0]

    @staticmethod
    async def _get(abstract_stock: Stock, from_datestamp: datetime.date, to_datestamp: datetime.date) -> List["StockMarketData"]:
        from_datestamp = common.get_next_business_day_adjusted_date(from_datestamp)
        to_datestamp = common.get_previous_business_day_adjusted_date(to_datestamp)
        from_timestamp: datetime.datetime = datetime.datetime.combine(from_datestamp, datetime.time.min, tzinfo=ZoneInfo("America/New_York"))
        to_timestamp: datetime.datetime = datetime.datetime.combine(to_datestamp, datetime.time.min, tzinfo=ZoneInfo("America/New_York"))
        stock = cast(IBKRStock, await IBKRStock._get_local_object(abstract_stock))
        raw_ohlc_data_list: List[Dict[str, float]] = await IBKRStockMarketData._get_raw_ohlc_data(stock.contract_id, from_timestamp, to_timestamp)

        return [IBKRStockMarketData(
            open=Decimal(str(ohlc_data["o"])),
            high=Decimal(str(ohlc_data["h"])),
            low=Decimal(str(ohlc_data["l"])),
            close=Decimal(str(ohlc_data["c"])),
            timestamp=datetime.datetime.fromtimestamp(ohlc_data["t"] / 1000),
            stock=stock)
            for ohlc_data in raw_ohlc_data_list
            if from_timestamp.timestamp() <= ohlc_data["t"] / 1000 <= to_timestamp.timestamp()]

    @staticmethod
    async def get_latest(abstract_stock: Stock) -> "IBKRStockMarketData":
        ibkr_stock: IBKRStock = cast(IBKRStock, await IBKRStock._get_local_object(abstract_stock))
        raw_data: Dict[str, Any] = await IBKRStockMarketData._get_raw_latest_ohlc_data(ibkr_stock.contract_id)
        open: Decimal = Decimal(raw_data["7295"].replace("H", "").replace("C", ""))
        high: Decimal = Decimal(raw_data["70"].replace("H", "").replace("C", ""))
        low: Decimal = Decimal(raw_data["71"].replace("H", "").replace("C", ""))
        close: Decimal = Decimal(raw_data["31"].replace("H", "").replace("C", ""))
        return IBKRStockMarketData(
            open=open,
            high=high,
            low=low,
            close=close,
            timestamp=datetime.datetime.now(),
            stock=ibkr_stock
        )


class IBKRPosition(DataClass):
    quantity: Decimal
    contract: SerializeAsAny[IBKRContract]
    average_cost: Decimal
    market_value: Decimal
    profit: Decimal

    @staticmethod
    @alru_cache(maxsize=50, ttl=60)  # 60 second cache
    async def _get_raw_all_position_data(account_id: str) -> List[Dict[str, Any]]:
        session: AsyncHTTPSession = await get_session()
        response: HTTPResponse = await session.get(f"{await get_base_url()}portfolio/{account_id}/positions/0")

        return list(filter(lambda d: float(d["position"]) != 0.0, response.data))

    @staticmethod
    async def get_all(account_id: str) -> List["IBKRPosition"]:
        positon_list: List[IBKRPosition] = []

        for data in await IBKRPosition._get_raw_all_position_data(account_id):
            quantity: Decimal = Decimal(data["position"])
            average_cost: Decimal = Decimal(data["avgCost"]) * quantity
            market_value: Decimal = Decimal("0")
            contract: IBKRContract = await IBKRContract.find_by_contract_id(data["conid"])

            if isinstance(contract, Stock):  # type: ignore[unreachable]
                stock_market_data: StockMarketData = await StockMarketData.get_latest(contract)  # type: ignore[unreachable]
                market_value = stock_market_data.close * quantity
            elif isinstance(contract, Option):  # type: ignore[unreachable]
                option_market_data: OptionMarketData = await OptionMarketData.get_latest(contract)  # type: ignore[unreachable]
                market_value = option_market_data.last * quantity * Decimal(100)

            positon_list.append(IBKRPosition(
                contract=contract,
                quantity=quantity,
                market_value=market_value,
                average_cost=average_cost,
                profit=market_value - average_cost
            ))

        return positon_list


class IBKRAccount(DataClass):
    id: str
    name: str
    position_list: List[IBKRPosition]
    cash_value: Decimal
    total_value: Decimal

    @staticmethod
    @alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
    async def _get_raw_all_account_list() -> List[Dict[str, Any]]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        response: HTTPResponse = await session.get(f"{base_url}portfolio/accounts")

        return list(filter(lambda d: d["id"][-1].isdigit(), response.data))

    @staticmethod
    @alru_cache(maxsize=50, ttl=60)  # 60 second cache
    async def _get_raw_all_account_summary_data(account_id: str) -> Dict[str, Any]:
        base_url: str = await get_base_url()
        session: AsyncHTTPSession = await get_session()
        response: HTTPResponse = await session.get(f"{base_url}iserver/account/{account_id}/summary")

        return response.data

    @staticmethod
    async def get_all() -> List["IBKRAccount"]:
        ibkr_account_list: List[IBKRAccount] = []

        for data in await IBKRAccount._get_raw_all_account_list():
            account_id: str = data["id"]
            raw_account_data: Dict[str, Any] = await IBKRAccount._get_raw_all_account_summary_data(account_id)
            cash_value: Decimal = Decimal(raw_account_data["totalCashValue"])
            position_list: List[IBKRPosition] = await IBKRPosition.get_all(account_id)
            # total_value: Decimal = sum(position.market_value for position in position_list) + cash_value
            total_value: Decimal = Decimal(raw_account_data["netLiquidationValue"])

            ibkr_account_list.append(
                IBKRAccount(
                    id=account_id,
                    name=data["accountAlias"] if data["accountAlias"] else "",
                    position_list=position_list,
                    cash_value=cash_value,
                    total_value=total_value
                )
            )

        return ibkr_account_list



import asyncio
import datetime
from abc import ABC, abstractmethod
from decimal import Decimal
from enum import Enum
from typing import List, Optional, Callable

import numpy
from async_lru import alru_cache
from pydantic import ConfigDict

from sirius.common import DataClass, Currency
from sirius.exceptions import SiriusException
from sirius.http_requests import ServerSideException


def calculate_annualized_return(starting: Decimal, ending: Decimal, number_of_days: int) -> Decimal:
    daily_return = (ending / starting)
    annualization_factor = Decimal("365") / Decimal(number_of_days)
    annualized_return = daily_return ** annualization_factor - Decimal("1")
    return annualized_return


class MarketDataException(SiriusException):
    pass


class Exchange(Enum):
    NASDAQ = "NASDAQ"
    NYSE = "NYSE"
    ARCA = "ARCA"


class OptionType(Enum):
    PUT = "PUT"
    CALL = "CALL"


class Asset(DataClass):
    name: str
    currency: Currency


class Stock(Asset, ABC):
    ticker: str

    @classmethod
    async def _get_local_object(cls, other: "Stock") -> "Stock":
        if isinstance(other, cls):
            return other
        else:
            stock: Stock | None = await cls._find(other.ticker)
            if not stock:
                raise MarketDataException(f"Could not find stock with ticker: {other.ticker}")

            return stock

    @staticmethod
    @abstractmethod
    async def _find(ticker: str) -> Optional["Stock"]:
        ...

    @staticmethod
    async def find(ticker: str) -> Optional["Stock"]:
        from sirius.market_data.massive import MassiveStock
        return await MassiveStock._find(ticker)

    @staticmethod
    async def get(ticker: str) -> "Stock":
        stock: Stock = await Stock.find(ticker)
        if not stock:
            raise MarketDataException(f"Could not find stock with ticker: {ticker}")

        return stock


class Option(Asset, ABC):
    id: str
    underlying_stock: Stock
    strike_price: Decimal
    expiry_date: datetime.date
    type: OptionType

    async def get_underlying_stock(self) -> Stock:
        return self.underlying_stock

    @classmethod
    async def _get_local_object(cls, other: "Option") -> "Option":
        if isinstance(other, cls):
            return other
        else:
            option: Option | None = await cls._find(other.id)
            if not option:
                raise MarketDataException(f"Could not find option with ID: {other.id}")

            return option

    @staticmethod
    @abstractmethod
    async def _find(option_id: str) -> Optional["Option"]:
        ...

    @staticmethod
    @abstractmethod
    async def _find_all_options(ticker: str, number_of_days_to_expiry: int) -> List["Option"]:
        ...

    @staticmethod
    def _get_id(ticker: str, expiry_date: datetime.date, strike_price: Decimal, option_type: OptionType) -> str:
        return f"{ticker}{expiry_date.strftime("%y%m%d")}{"C" if option_type == OptionType.CALL else "P"}{str(int(strike_price * 1000)).zfill(8)}"

    @staticmethod
    async def find_all_options(ticker: str, number_of_days_to_expiry: int) -> List["Option"]:
        from sirius.market_data.massive import MassiveOption
        return await MassiveOption._find_all_options(ticker, number_of_days_to_expiry)


class StockMarketData(DataClass, ABC):
    open: Decimal
    high: Decimal
    low: Decimal
    close: Decimal
    timestamp: datetime.datetime
    stock: Stock

    @staticmethod
    @abstractmethod
    async def _get(abstract_stock: Stock, from_datestamp: datetime.date, to_datestamp: datetime.date) -> List["StockMarketData"]:
        ...

    @staticmethod
    async def get(stock: Stock, from_datestamp: datetime.date, to_datestamp: datetime.date | None = None) -> List["StockMarketData"]:
        to_datestamp = datetime.date.today() if not to_datestamp else to_datestamp
        from sirius.market_data.massive import MassiveStockMarketData
        return await MassiveStockMarketData._get(stock, from_datestamp, to_datestamp)

    @staticmethod
    async def get_latest(stock: Stock) -> "StockMarketData":
        from sirius.market_data.massive import MassiveStockMarketData
        return await MassiveStockMarketData.get_latest(stock)


class OptionMarketData(DataClass, ABC):
    option: Option
    last: Decimal

    async def get_option(self) -> Option:
        return self.option

    @staticmethod
    @abstractmethod
    async def _get_latest(option: Option) -> "OptionMarketData":
        ...

    @staticmethod
    async def get_latest(option: Option) -> "OptionMarketData":
        from sirius.market_data.massive import MassiveOptionMarketData
        return await MassiveOptionMarketData._get_latest(option)


class StockPerformance(DataClass):
    position_open: StockMarketData
    position_close: StockMarketData
    absolute_return: Decimal
    annualized_return: Decimal
    model_config = ConfigDict(frozen=True)

    @staticmethod
    def _construct(position_open: StockMarketData, position_close: StockMarketData) -> "StockPerformance":
        open_price: Decimal = position_open.open if position_open == position_close else position_open.close
        absolute_return: Decimal = (position_close.close - open_price) / open_price
        number_of_days = (position_close.timestamp.date() - position_open.timestamp.date()).days
        annualized_return = calculate_annualized_return(open_price, position_close.close, max(1, number_of_days))

        return StockPerformance(
            position_open=position_open,
            position_close=position_close,
            absolute_return=absolute_return,
            annualized_return=annualized_return
        )

    @staticmethod
    @alru_cache(maxsize=50, ttl=600)  # 5 min cache
    async def get(ticker: str, from_datestamp: datetime.date, to_datestamp: datetime.date) -> Optional["StockPerformance"]:
        stock: Stock = await Stock.get(ticker)
        market_data_list: List[StockMarketData] = await StockMarketData.get(stock, from_datestamp, to_datestamp)

        if not market_data_list or (from_datestamp != to_datestamp and len(market_data_list) <= 1):
            return None

        position_open: StockMarketData = min(market_data_list, key=lambda m: m.timestamp)
        position_close: StockMarketData = max(market_data_list, key=lambda m: m.timestamp)
        return StockPerformance._construct(position_open, position_close)


class StockPerformanceAnalysis(DataClass):
    stock_performance_list: List[StockPerformance]
    standard_deviation_absolute_return: Decimal
    standard_deviation_annualized_return: Decimal
    mean_absolute_return: Decimal
    mean_annualized_return: Decimal
    median_absolute_return: Decimal
    median_annualized_return: Decimal
    min_absolute_return: Decimal
    min_annualized_return: Decimal
    max_absolute_return: Decimal
    max_annualized_return: Decimal

    async def get_probability_of_stock_reaching_target(self, stock: Stock, target_price: Decimal) -> Decimal:
        latest_price: StockMarketData = await StockMarketData.get_latest(stock)
        absolute_return: Decimal = (target_price - latest_price.close) / latest_price.close
        stock_performance_list: List[StockPerformance] = self.stock_performance_list
        number_of_stock_performance_reaching_target: int = len([s for s in stock_performance_list if s.absolute_return >= absolute_return])

        return Decimal(number_of_stock_performance_reaching_target) / Decimal(len(stock_performance_list))

    @staticmethod
    @alru_cache(maxsize=50, ttl=600)  # 5 min cache
    async def get(ticker: str, number_of_days_invested: int, number_of_days_to_analyse: int | None = None, analysis_end_datestamp: datetime.date | None = None) -> "StockPerformanceAnalysis":
        today: datetime.date = datetime.datetime.now().date()
        number_of_days_to_analyse = (today - today.replace(year=today.year - 10)).days if not number_of_days_to_analyse else number_of_days_to_analyse
        analysis_end_datestamp = datetime.date.today() if not analysis_end_datestamp else analysis_end_datestamp
        analysis_dates_list: List[datetime.date] = [analysis_end_datestamp - datetime.timedelta(days=day + number_of_days_invested) for day in range(number_of_days_to_analyse)]
        raw_stock_performance_list: List[Optional[StockPerformance]] = list(await asyncio.gather(*[StockPerformance.get(ticker, date, date + datetime.timedelta(days=number_of_days_invested)) for date in analysis_dates_list]))
        stock_performance_list: List[StockPerformance] = [result for result in raw_stock_performance_list if result is not None]

        if not stock_performance_list:
            raise ServerSideException("Did not retrieve any Market Data")

        absolute_return_list: List[Decimal] = [stock_performance.absolute_return for stock_performance in stock_performance_list]
        annualized_return_list: List[Decimal] = [stock_performance.annualized_return for stock_performance in stock_performance_list]
        standard_deviation_absolute_return: Decimal = Decimal(str(numpy.std(absolute_return_list)))  # type: ignore[arg-type]
        standard_deviation_annualized_return: Decimal = Decimal(str(numpy.std(annualized_return_list)))  # type: ignore[arg-type]
        mean_absolute_return: Decimal = Decimal(str(numpy.mean(absolute_return_list)))  # type: ignore[arg-type]
        mean_annualized_return: Decimal = Decimal(str(numpy.mean(annualized_return_list)))  # type: ignore[arg-type]
        median_absolute_return: Decimal = Decimal(str(numpy.median(absolute_return_list)))  # type: ignore[arg-type]
        median_annualized_return: Decimal = Decimal(str(numpy.median(annualized_return_list)))  # type: ignore[arg-type]
        min_absolute_return: Decimal = Decimal(str(numpy.min(absolute_return_list)))  # type: ignore[arg-type]
        min_annualized_return: Decimal = Decimal(str(numpy.min(annualized_return_list)))  # type: ignore[arg-type]
        max_absolute_return: Decimal = Decimal(str(numpy.max(absolute_return_list)))  # type: ignore[arg-type]
        max_annualized_return: Decimal = Decimal(str(numpy.max(annualized_return_list)))  # type: ignore[arg-type]

        return StockPerformanceAnalysis(
            stock_performance_list=stock_performance_list,
            standard_deviation_absolute_return=standard_deviation_absolute_return,
            standard_deviation_annualized_return=standard_deviation_annualized_return,
            mean_absolute_return=mean_absolute_return,
            mean_annualized_return=mean_annualized_return,
            median_absolute_return=median_absolute_return,
            median_annualized_return=median_annualized_return,
            min_absolute_return=min_absolute_return,
            min_annualized_return=min_annualized_return,
            max_absolute_return=max_absolute_return,
            max_annualized_return=max_annualized_return
        )


class OptionPerformanceAnalysis(DataClass):
    option: Option
    stock_performance_analysis: StockPerformanceAnalysis
    itm_probability: Decimal

    @staticmethod
    async def get(option_contract: Option) -> "OptionPerformanceAnalysis":
        number_of_days_invested: int = (option_contract.expiry_date - datetime.datetime.now().date()).days
        underlying_stock: Stock = await option_contract.get_underlying_stock()
        stock_performance_analysis: StockPerformanceAnalysis = await StockPerformanceAnalysis.get(underlying_stock.ticker, number_of_days_invested)
        itm_probability: Decimal = await stock_performance_analysis.get_probability_of_stock_reaching_target(underlying_stock, option_contract.strike_price)

        return OptionPerformanceAnalysis(
            option=option_contract,
            stock_performance_analysis=stock_performance_analysis,
            itm_probability=itm_probability
        )



import os
from typing import List

from setuptools import setup, find_packages


def get_package_name() -> str:
    return "aorta_sirius" if is_production_branch() else "aorta_sirius_dev"


def is_installing() -> bool:
    return os.path.isfile("PKG-INFO")


def get_version() -> str:
    if is_installing():
        with open("PKG-INFO", "r") as package_info_file:
            return package_info_file.read().split("\n")[2].replace("Version: ", "")

    import requests
    from requests import Response

    response: Response = requests.get(f"https://pypi.org/pypi/{get_package_name()}/json")
    assert response.status_code == 200
    version_number: str = response.json()["info"]["version"]
    next_subversion_number: int = int(version_number.split(".")[-1]) + 1
    return ".".join(version_number.split(".")[0:-1], ) + "." + str(next_subversion_number)


def is_production_branch() -> bool:
    return "production" == os.getenv("GITHUB_REF_NAME") or os.path.isdir("src/aorta_sirius.egg-info")


def get_required_packages() -> List[str]:
    requirements_file_path: str = "requirements.txt" if os.path.isfile("requirements.txt") else f"src/{get_package_name()}.egg-info/requires.txt"

    with open(requirements_file_path, "r") as requirements_file:
        return requirements_file.read().split("\n")


setup(
    name=get_package_name(),
    version=get_version(),
    url="https://github.com/kontinuum-investments/Aorta-Sirius",
    author="Kavindu Athaudha",
    author_email="kavindu@kih.com.sg",
    packages=find_packages(where="src", include=["sirius*"]),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=get_required_packages()
)



import datetime
import time
from decimal import Decimal
from typing import List, Dict, Any, Optional
from zoneinfo import ZoneInfo

from async_lru import alru_cache

from sirius import common
from sirius.common import Currency
from sirius.constants import SiriusEnvironmentSecretKey
from sirius.http_requests import AsyncHTTPSession, HTTPResponse
from sirius.market_data import Option, Stock, OptionType, OptionMarketData, StockMarketData

BASE_URL: str = "https://www.alphavantage.co/query"


class AlphaVantageOption(Option):

    @staticmethod
    @alru_cache(maxsize=50, ttl=43_200)  # 12 hour cache
    async def _get_raw_data(ticker: str) -> Dict[Any, Any]:
        api_key: str = await common.get_environmental_secret(SiriusEnvironmentSecretKey.ALPHA_VANTAGE_API_KEY)
        response: HTTPResponse = await AsyncHTTPSession(BASE_URL).get(f"{BASE_URL}", query_params={
            "function": "HISTORICAL_OPTIONS",
            "symbol": ticker,
            "apikey": api_key,
            "entitlement": "delayed"
        })

        return response.data

    @staticmethod
    async def _find(option_id: str) -> Optional["Option"]:
        ticker: str = option_id.split(" | ")[0]
        option_list: List[AlphaVantageOption] = await AlphaVantageOption._find_all_from_ticker(ticker)
        option_map: Dict[str, AlphaVantageOption] = {option.id: option for option in option_list}
        return option_map.get(option_id)

    @staticmethod
    async def _find_all_from_ticker(ticker: str) -> List["AlphaVantageOption"]:
        option_list: List[AlphaVantageOption] = []
        underlying_stock: Stock = await Stock.find(ticker)
        if not underlying_stock:
            return []

        raw_data: Dict[Any, Any] = await AlphaVantageOption._get_raw_data(ticker)
        for data in raw_data["data"]:
            expiry_date: datetime.date = datetime.datetime.strptime(data["expiration"], "%Y-%m-%d").date()
            if expiry_date < datetime.datetime.now().date():
                continue

            strike_price: Decimal = Decimal(data["strike"])
            option_type: OptionType = OptionType(str(data["type"]).upper())
            option_id: str = Option._get_id(ticker, expiry_date, strike_price, option_type)
            option_list.append(AlphaVantageOption(
                id=option_id,
                underlying_stock=underlying_stock,
                strike_price=strike_price,
                expiry_date=expiry_date,
                type=option_type,
                currency=underlying_stock.currency,
                name=option_id
            ))

        return option_list

    @staticmethod
    async def _find_all_options(ticker: str, number_of_days_to_expiry: int) -> List["Option"]:
        data = await AlphaVantageOption._find_all_from_ticker(ticker)
        return [option for option in data
                if (option.expiry_date - datetime.datetime.now().date()).days == number_of_days_to_expiry]


class AlphaVantageOptionMarketData(OptionMarketData):

    @staticmethod
    @alru_cache(maxsize=50, ttl=43_200)  # 12 hour cache
    async def _get_all_last_map(ticker: str) -> Dict[str, Decimal]:
        all_last_map: Dict[str, Decimal] = {}
        raw_data: Dict[Any, Any] = await AlphaVantageOption._get_raw_data(ticker)
        for data in raw_data["data"]:
            expiry_date: datetime.date = datetime.datetime.strptime(data["expiration"], "%Y-%m-%d").date()
            strike_price: Decimal = Decimal(data["strike"])
            option_type: OptionType = OptionType(str(data["type"]).upper())
            option_id: str = Option._get_id(ticker, expiry_date, strike_price, option_type)
            all_last_map[option_id] = Decimal(data["last"])

        return all_last_map

    @staticmethod
    async def _get_latest(abstract_option: Option) -> "AlphaVantageOptionMarketData":
        underlying_stock: Stock = await abstract_option.get_underlying_stock()
        all_map: Dict[str, Decimal] = await AlphaVantageOptionMarketData._get_all_last_map(underlying_stock.ticker)
        last: Decimal = all_map[abstract_option.id]
        return AlphaVantageOptionMarketData(option=abstract_option, last=last)


class AlphaVantageStockMarketData(StockMarketData):

    @staticmethod
    @alru_cache(maxsize=50, ttl=43_200)  # 12 hour cache
    async def _get_raw_time_series_data(ticker: str) -> Dict[datetime.date, Any]:
        api_key: str = await common.get_environmental_secret(SiriusEnvironmentSecretKey.ALPHA_VANTAGE_API_KEY)
        response: HTTPResponse = await AsyncHTTPSession(BASE_URL).get(f"{BASE_URL}", query_params={
            "function": "TIME_SERIES_DAILY",
            "symbol": ticker,
            "outputsize": "full",
            "apikey": api_key,
            "entitlement": "delayed"
        })

        return {
            datetime.date.fromisoformat(date_str): value
            for date_str, value in response.data["Time Series (Daily)"].items()
        }

    @staticmethod
    @alru_cache(maxsize=50, ttl=60)  # 60 second cache
    async def _get_raw_latest_stock_market_data(ticker: str) -> Dict[Any, Any]:
        api_key: str = await common.get_environmental_secret(SiriusEnvironmentSecretKey.ALPHA_VANTAGE_API_KEY)
        response: HTTPResponse = await AsyncHTTPSession(BASE_URL).get(f"{BASE_URL}", query_params={
            "function": "GLOBAL_QUOTE",
            "symbol": ticker,
            "apikey": api_key,
            "entitlement": "delayed"
        })

        return response.data["Global Quote - DATA DELAYED BY 15 MINUTES"]

    @staticmethod
    async def _get(abstract_stock: Stock, from_datestamp: datetime.date, to_datestamp: datetime.date) -> List["StockMarketData"]:
        raw_data: Dict[datetime.date, Any] = await AlphaVantageStockMarketData._get_raw_time_series_data(abstract_stock.ticker)
        return [AlphaVantageStockMarketData(
            open=Decimal(raw_data[datestamp]["1. open"]),
            high=Decimal(raw_data[datestamp]["2. high"]),
            low=Decimal(raw_data[datestamp]["3. low"]),
            close=Decimal(raw_data[datestamp]["4. close"]),
            timestamp=datetime.datetime.combine(datestamp, datetime.time.min, tzinfo=ZoneInfo("America/New_York")),
            stock=abstract_stock
        ) for datestamp in raw_data.keys()
            if from_datestamp <= datestamp <= to_datestamp]

    @staticmethod
    async def get_latest(abstract_stock: Stock) -> "AlphaVantageStockMarketData":
        raw_data: Dict[Any, Any] = await AlphaVantageStockMarketData._get_raw_latest_stock_market_data(abstract_stock.ticker)

        return AlphaVantageStockMarketData(
            open=Decimal(raw_data["02. open"]),
            high=Decimal(raw_data["03. high"]),
            low=Decimal(raw_data["04. low"]),
            close=Decimal(raw_data["05. price"]),
            timestamp=datetime.datetime.strptime(raw_data["07. latest trading day"], "%Y-%m-%d"),
            stock=abstract_stock
        )


class AlphaVantageStock(Stock):

    @staticmethod
    @alru_cache(maxsize=50, ttl=86_400)  # 24 hour cache
    async def _get_raw_ticker_search_data(ticker: str) -> List[Dict[str, Any]]:
        api_key: str = await common.get_environmental_secret(SiriusEnvironmentSecretKey.ALPHA_VANTAGE_API_KEY)
        response: HTTPResponse = await AsyncHTTPSession(BASE_URL).get(f"{BASE_URL}", query_params={
            "function": "SYMBOL_SEARCH",
            "keywords": ticker,
            "apikey": api_key
        })

        return response.data["bestMatches"]

    @staticmethod
    async def _find(ticker: str) -> Optional["Stock"]:
        raw_data_list: List[Dict[str, Any]] = await AlphaVantageStock._get_raw_ticker_search_data(ticker)
        try:
            raw_data: Dict[str, Any] = max((raw_data for raw_data in raw_data_list if raw_data["1. symbol"] == ticker and raw_data["4. region"] == "United States" and raw_data["8. currency"] == "USD"), key=lambda r: float(r["9. matchScore"]))
        except ValueError:
            return None

        return AlphaVantageStock(
            name=raw_data['2. name'],
            currency=Currency(raw_data['8. currency']),
            ticker=raw_data['1. symbol']
        )
